# $Id: menuHelp.tcl,v 2.12 2005-04-20 09:51:36-04 m1tca00 Exp $
#
#   Menu with help and search widget.
#
# .SH "SYNOPSIS"
#   *package require menuHelp*
#   .br
#   *menuHelp* $objName \fI[CLASS OPTIONS]\fR
#   .br
#   $objName *method* \fI[METHOD OPTIONS]\fR
#
# .SH "DESCRIPTION"
#   Converts a menu into a tixTree with a help & search facility.
#
# $Author: m1tca00 $

# $Locker:  $
#
# $Log: menuHelp.tcl,v $
# Revision 2.12  2005-04-20 09:51:36-04  m1tca00
# Bind <Enter> to glimpseSearch.  If no glimpse directory exists,
#   glimpseSearch now falls back to apropos.
#
# Revision 2.11  2004/12/05 18:45:48  m1tca00
# Glimpse search will assume only one argument and it must begin with
#   a hyphen.
#
# Revision 2.10  2004/11/13 03:23:06  m1tca00
# Improved glimpse search.
#
# Revision 2.9  2004/08/09 21:18:24  m1tca00
# Add search string support to xvi (same as gvim).
#
# Revision 2.8  2004/08/02 13:36:52  m1tca00
# Use /bks/proj/bin for glimpse.
#
# Revision 2.7  2004-07-19 15:20:12-04  m1tca00
# Allow double spacing but not triple spacing.
#
# Revision 2.6  2004-07-19 15:17:06-04  m1tca00
# Don't count lines with start with dot (.) in readHelpFile.
#
# Revision 2.5  2004-07-19 14:27:03-04  m1tca00
# Documentation cleanup.
# readHelpFile will deal with man page comments.
#
# Revision 2.4  2003-09-10 15:49:12-04  m1tca00
# Added WheelMouse support.
#
# Revision 2.3  2003-08-11 15:53:43-04  m1tca00
# Index numbers in glimpseindex.
#
# Revision 2.2  2003-05-14 12:02:31-04  m1tca00
# imageLib images moved to imageLib namespace.
#
# Revision 2.1  2001-11-19 17:15:53-05  m1tca00
# Since glimpseOut is a child of top (in 2.0), it is no longer necessary
#   to destroy glimpseOut separately.
#
# Revision 2.0  2001-11-19 16:15:28-05  m1tca00
# Replaced stooop implementation with [incr Tcl].
#

package provide menuHelp 2.0

# Required extensions
package require Tclx
package require Tix
package require BWidget
package require Itcl
package require OutWindow
package require imageLib

itcl::class menuHelp {
    ####################################################################
    # Configuration options                                            #
    ####################################################################

    # Title of menu
    public variable title  {} {
	if {[info exists status]} {
	    $status configure -text "$title\nMenu"
	}
    }

    # Name of menu which we are providing help for
    public variable menu

    # Directory where glimpse index lives
    public variable glimpseDir  {}

    # Help loaded flag
    public variable helpLoaded 0

    ####################################################################
    # Private variables                                                #
    ####################################################################

    # Version (set by RCS)
    private variable version {$Revision: 2.12 $}

    # GUI components
    private variable top
    private variable buttons
    private variable status
    private variable help
    private variable entry
    private variable tree
    private variable glimpseOut
    private variable disabledStyle

    # Tree array
    #	Indexed by:
    #	    $key,label	    Label
    #	    $key,cmd	    Command
    #	    $key,helpText   Help text
    #	    $key,editFile   File available for editing (if any)
    #	    $key,helpFile   File to get help from (if any)
    private variable _tree
    # Size of tree
    private variable menuCnt 0
    # Meter count
    private variable meter
    # Time latest file was modified
    private variable helpMtime
    # Name of latest modified file
    private variable helpMtimeFileName

    ####################################################################
    # Constructor/Destructor                                           #
    ####################################################################

    constructor {args} {}
    destructor {}

    # menuHelp::version
    #
    #	    Return the version number.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The RCS version number is returned.
    public method version {}

    ####################################################################
    # User preferences                                                 #
    ####################################################################

    # menuHelp::preferences
    #
    #	    Initialize preferences.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Default resources are set and options user resources are
    #	    read from "~/.menuhelp-defaults".
    private method preferences {}

    # menuHelp::changeEditor
    #
    #	    Allow user to change the editor.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #       The user is prompted to enter the name of the editing
    #       command that they would like to use.  Default can be changed
    #       in "~/.menuhelp-defaults".  This method is bound to the key
    #       "E" in the widget.
    public method changeEditor {}

    # menuHelp::curSelection
    #
    #	    Get the current selection in the tree (if any).
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The current selection in $tree (if any) is returned.
    public method curSelection {}

    ####################################################################
    # Static Procedures                                                #
    ####################################################################

    # menuHelp::readHelpFile
    #
    #	    Read the comments from the fileName.
    #
    # Arguments:
    #	    fileName	    Name of editFile or helpFile.
    #
    # Results:
    #       The first 21 lines (or lines before the first non-blank
    #       line) will be read from $fileName.  The first line will
    #       always be ignored.  Comment characters (based on filetype)
    #       will be removed.  These comments will be formatted and
    #       returned.
    proc readHelpFile {fileName} {}

    # menuHelp::bindButton
    #
    #	    Find which entry was clicked.
    #
    # Arguments:
    #	    w		widget.
    #	    x		x coordinate.
    #	    y		y coordinate.
    #
    # Results:
    #       Tix doesn't offer bindings for button-2 or -3 clicks, so we
    #       use Tix's method to see what item is clicked.  This method
    #       will be bound to button-2 and -3 clicks in the $tree.  It
    #       will return the $key to the item nearest the mouse (if any).
    proc bindButton {w x y} {}

    ####################################################################
    # Methods for building the tree                                    #
    ####################################################################

    # menuHelp::buildTree
    #
    #	    Loop through menu antries and add to the tree.
    #
    # Arguments:
    #	    addPath	Root path to add menu to.
    #	    m		Menu.
    #
    # Results:
    #       Loop over all the entries of a given menu ($m) and all
    #       entries are added to the tree below $addPath.  If an entry
    #       is a cascading menu, buildTree will be called again with
    #       the cascading menu (as $m) and the new current path (as
    #       $addPath). $menuCnt is incremented for each entry to keep
    #       track of the size of the tree.  Only the following entry
    #       types are added (anything else is ignored): command,
    #       checkbutton, cascade & separator.  Nothing is returned.
    private method buildTree {addPath m}

    # menuHelp::getComments
    #
    #	    Extract comments from a command.
    #
    # Arguments:
    #	    cmd		Command.
    #
    # Results:
    #       Comments of the command ($cmd) will be returned (if any).
    private method getComments {cmd}

    # menuHelp::assignAttributes
    #
    #       Assign the help file and the edit file for key.
    #
    # Arguments:
    #	    key		Menu entry key.
    #
    # Results:
    #	    Check if there is an editFile and helpFile for the $key
    #	    command.  The helpText will be set by stripping comment
    #	    characters from the comments.  Nothing will be returned.
    #
    #       An editFile exists by adding the "Edit file: _filename_"
    #       as the first comment of the command.  A helpFile exists by
    #       adding "Help file: _filename_" as the second comment of the
    #       command.  A helpFile can only exist if there is an editFile.
    #       If there is an editFile but no helpFile, editFile will be
    #       used as the helpFile.
    private method assignAttributes {key}

    ####################################################################
    # Methods for hiding or displaying branches                        #
    ####################################################################

    # menuHelp::hideAll
    #
    #	    Hide all entries.
    #
    # Arguments:
    #	    key		Hide everything below this leaf (optional).
    #
    # Results:
    #	    All children of $key will be hidden (branches collapsed).
    #	    Nothing will be returned.
    private method hideAll {{key {}}}

    # menuHelp::findPath
    #
    #	    Return all keys that lead to path.
    #
    # Arguments:
    #	    key		Menu entry key.
    #
    # Results:
    #	    All immediate ancestors of $key will be returned as a list.
    private method findPath {key}

    # menuHelp::showKey
    #
    #	    Show all keys leading to child.
    #
    # Arguments:
    #	    key		Last key we want to see.
    #
    # Results:
    #	    All branches leading to $key will be uncollapsed in the
    #	    $tree widget.  Nothing will be returned.
    private method showKey {key}

    ####################################################################
    # Methods for executing or editing menu items                      #
    ####################################################################

    # menuHelp::showCmd
    #
    #	    Show the command of a menu entry.
    #
    # Arguments:
    #	    key		Menu entry key.
    #
    # Results:
    #       If $key has a command associated with it, the command will
    #       be shown in the $help frame.  This method is bound to
    #       button-2 in the widget.  Nothing will be returned.
    private method showCmd {key}

    # menuHelp::runEntry
    #
    #	    Execute menu entry.
    #
    # Arguments:
    #	    key		Menu entry key.
    #
    # Results:
    #	    If $key has a command associated with it, the command will
    #	    be executed and the $help frame will be updated to indicate
    #	    so.  Nothing will be returned.
    private method runEntry {key}

    # menuHelp::editEntry
    #
    #	    Edit the editFile for a menu entry.
    #
    # Arguments:
    #	    key		Menu entry key (optional).
    #
    # Results:
    #	    If $key is non-null, it will be set to be the current
    #	    selection.  If $key has an editFile associated with it, the
    #	    editFile will be edited with the current editor.  Nothing
    #	    will be returned.
    private method editEntry {{key {}}}

    ####################################################################
    # Methods for the HELP/SEARCH facility                             #
    ####################################################################

    # menuHelp::searchHelp
    #
    #	    Show the SEARCH help.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The text of the SEARCH help will be displayed in the $help
    #	    frame.  Nothing will be returned.
    private method searchHelp {}

    # menuHelp::getHelp
    #
    #	    Get helpText for specified key.
    #
    # Arguments:
    #	    key		Menu entry key.
    #
    # Results:
    #	    If not already loaded, the helpText will be set for $key.
    #	    Nothing will be returned.
    private method getHelp {key}

    # menuHelp::reloadHelp
    #
    #	    Reload the helpText.
    #
    # Arguments:
    #	    key		Menu entry key.
    #
    # Results:
    #	    If there is a helpFile associated with $key, the helpText
    #	    will be reloaded (even if it had previosuly been loaded).
    private method reloadHelp {key}

    # menuHelp::showHelp
    #
    #	    Show the helpText.
    #
    # Arguments:
    #	    key		Menu entry key (optional).
    #	    helpText	Text to show in $help frame (optional).
    #
    # Results:
    #	    If $key is non-null, the helpText associated with that key
    #	    will be displayed in the $help frame.  If $key is null, the
    #	    $helpText passed will be displayed instead.  If $helpText
    #	    (either from the key or passed as an argument) are null, the
    #	    default helpText will be displayed.  Nothing will be
    #	    returned.
    public method showHelp {{key {}} {helpText {}}}

    # menuHelp::loadAllHelp
    #
    #	    Load ALL the helpText (needed for full search).
    #
    # Arguments:
    #	    key		Locate in $tree to begin searching (optional).
    #
    # Results:
    #	    Every child of $key will have the helpText loaded.  This
    #	    method will be called recursively down the tree.  Nothing
    #	    will be returned.
    private method loadAllHelp {{key .}}

    # menuHelp::updateMeter
    #
    #	    Update tixMeter.
    #
    # Arguments:
    #	    title	Title of meter widget.
    #
    # Results:
    #	    If $meter does not exist, a new tixMeter will be created and
    #	    $meter will be set to 0.  If $meter does exist, it will be
    #	    incremented and the tixMeter will be updated to the new
    #	    ratio of $meter/$menuCnt.  Nothing will be returned.
    private method updateMeter {{title Loading...}}

    # menuHelp::apropos
    #
    #	    Search help files and menu entries for regular expression.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    If helpText has not been loaded (!$helpLoaded), user will be
    #	    prompted if they want to load it now.  If yes, helpText will
    #	    be loaded.  If not, only already existing helpText will be
    #	    searched.  All loaded helpText and labels will be searched
    #	    for the regular expression in the $entry widget.  If any
    #	    keys match, all branches except those that match will be
    #	    hidden, otherwise a warning message will appear and the tree
    #	    display will not change.  Nothing will be returned.
    private method apropos {}

    # menuHelp::egrep
    #
    #       Search contents of help files for regular expression via
    #       egrep.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #       Execute egrep over every helpFile (if any) to search for
    #       the contents of the $entry widget.  If any keys match, all
    #       branches except those that match will be hidden, otherwise
    #       a warning message will appear and the tree display will not
    #       change.  Nothing will be returned.
    private method egrep {}

    # menuHelp::listHelpFiles
    #
    #	    Return a list of all helpFiles.
    #
    # Arguments:
    #	    addKeys	Also return associated keys (optional).
    #
    # Results:
    #	    The filenames of all helpFiles will be returned.  If
    #	    $addKeys, the associated keys will also be returned.
    public method listHelpFiles {{showKeys 0}}

    # menuHelp::glimpseOutdated
    #
    #       Check if the glimpse index is out-dated (it exists, but not
    #       up-to-date).
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #       The $glimpseIndex will be compared to all helpFiles.  Return
    #       0 if it is OK, 1 if it needs to be updated.
    private method glimpseOutdated {}

    # menuHelp::glimpseMtime
    #
    #	    Report the modification time of the glimpse index.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #       Display the time the glimpse index was updated.  If the
    #       index is out of date, indicate the time in red.  Return 0 on
    #       success, or 1 on failure to determine the time of the index.
    private method glimpseMtime {}

    # menuHelp::helpMtime
    #
    #	    Show the last modified time.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #       Show the name and time last modified of the most recently
    #       edited helpFile in the $help widget.  Return 0 on success,
    #       or 1 on failure.
    private method helpMtime {}

    # menuHelp::glimpseIndex
    #
    #	    Create a glimpse index of the helpFiles in user-specified
    #	    directory.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The glimpse index will be built.  Return 0 on success, or 1
    #	    on failure.
    public method glimpseIndex {}

    # menuHelp::glimpseSearch
    #
    #	    Use glimpse to search.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #       Search the glimpse database for the contents of the $entry
    #       widget.  If any keys match, all branches except those that
    #       match will be hidden and the glimpse results will be shown
    #       in an outputWindow, otherwise a warning message will appear
    #       and the tree display will not change.  Return 0 on success,
    #       or 1 on failure.
    private method glimpseSearch {}

    # menuHelp::glimpseHelp
    #
    #	    Give some help for Glimpse.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The glimpse help will be shown in the $help widget.  Nothing
    #	    will be returned.
    public method glimpseHelp {}
}

####################################################################
#                                                                  #
# Constructor/Destructor                                           #
#                                                                  #
####################################################################

itcl::body menuHelp::constructor {args} {
    # Configure public variables
    eval configure $args

    # $menu must be set, and it must be a menu
    if {![string match [info commands $menu] $menu]} {
	error "$menu does not exist"
    }

    # Create toplevel
    set top .menuHelp_[namespace tail $this]
    destroy $top
    toplevel $top -relief ridge -class Menuhelp
    wm title $top "Help Menu"

    # Get preferences from resources
    preferences

    # Buttons frame
    set buttons [frame $top.buttons -relief raised -borderwidth 4]
    set b $buttons
    button $b.exit -text Exit -relief raised -takefocus 0 \
	-command "itcl::delete object $this"
    button $b.search -text Search -relief raised -takefocus 0 \
	-command [itcl::code $this apropos]
    set status $b.status
    label $status -text "$title\nMenu" -foreground darkgreen -padx 3 \
	-justify left

    # Try and configure images created by imageLib
    catch "imageLib::imageConfig $b.exit Exit -relief flat"
    catch "imageLib::imageConfig $b.search Search -relief flat"

    # Glimpse index (will not be packed if no index exists...)
    set g [frame $top.glimpse -relief flat -borderwidth 0]
    button $g.search -text Glimpse -relief raised -takefocus 0 \
	-command [itcl::code $this glimpseSearch]
    menubutton $g.help -text Help -relief raised -takefocus 0 \
	-menu $g.help.m

    # Glimpse help menu
    set helpMenu [menu $g.help.m]
    $helpMenu add command -label "Brief Overview" \
	-command [itcl::code $this glimpseHelp]
    $helpMenu add command -label "Glimpse Manual" \
	-command "system {/bks/proj/bin/tkman.remote glimpse} &"
    $helpMenu add separator
    $helpMenu add command -label "Location of Index" \
	-command [itcl::code $this showHelp {} {\n$glimpseDir}]
    $helpMenu add command -label "Last File Updated" \
	-command [itcl::code $this helpMtime]
    $helpMenu add command -label "Update Index" \
	-command [itcl::code $this glimpseIndex]

    # Glimpse index modification time
    label $g.mtime -text "\n" -foreground darkgreen -padx 3 -justify left
    catch "imageLib::imageConfig $g.search Glimpse -relief flat"
    catch "imageLib::imageConfig $g.help Help -relief flat"

    # Create search entry
    tixLabelEntry $top.search -relief flat \
	-label "Enter search expression:" -labelside top
    set entry [$top.search subwidget entry]

    # Help text
    tixScrolledText $top.help -scrollbar y
    set help [$top.help subwidget text]
    $help configure -cursor {} -width 30 -height 9 -relief ridge \
	-takefocus 0 -state disabled -wrap word
    set rgb [winfo rgb $top [$top cget -background]]
    $help configure -background [format "#%04x%04x%04x" \
	    [expr {int([lindex $rgb 0] * .9)}] \
	    [expr {int([lindex $rgb 1] * .9)}] \
	    [expr {int([lindex $rgb 2] * .9)}]] \
	-foreground black

    # Menu tree
    tixTree $top.tree -command [itcl::code $this runEntry] \
	-browsecmd [itcl::code $this showHelp] \
	-options "
	    hlist.indicator	1
	    hlist.padX		3
	    hlist.padY		0
	    hlist.header	0
	    hlist.width 	30
	    hlist.height	17
	    hlist.separator	{.}
	    hlist.wideSelection 0
	    hlist.selectMode	browse
	    hlist.indent	25
	    hlist.foreground	blue
	"
    $top.tree configure -highlightcolor blue
    set tree [$top.tree subwidget hlist]
    # Configure disabled style
    set disabledStyle \
	[tixDisplayStyle text -refwindow $tree -foreground \
	    [option get $top disabledForeground DisabledForeground]]

    # Bindings
    bind $top <Destroy>		"itcl::delete object $this"
    bind $top <Control-q>	"itcl::delete object $this"
    bind $top.search <Enter>	"[itcl::code $this searchHelp] ; focus $entry"
    bind $b.search <Enter>	"[itcl::code $this searchHelp] ; focus $entry"
    bind $b.search <Button-2>	[itcl::code $this egrep]
    bind $g.search <Button-2>	[itcl::code $this glimpseIndex]
    bind $g.help <Button-2>	[itcl::code $this helpMtime]
    bind $b.status <Enter>	[itcl::code $this showHelp]
    bind $entry <Return>	[itcl::code $this glimpseSearch]
    bind $entry <Key-Next>	"$tree yview scroll 1 pages"
    bind $entry <Key-Prior>	"$tree yview scroll -1 pages"
    bind $entry <Key-Up>	"focus $tree ; tixHList:UpDown $tree prev {}"
    bind $entry <Key-Down>	"focus $tree ; tixHList:UpDown $tree next {}"
    bind $top.tree <Enter>	"focus $tree"
    bind $tree <Tab>		"focus $entry"
    bind $tree <Enter>		"[itcl::code $this showHelp] \
					\[[itcl::code $this curSelection]\]"
    bind $tree <Key-Home>	"$tree yview moveto 0"
    bind $tree <Key-End>	"$tree yview moveto 1"
    bind $tree <Key-e>		"[itcl::code $this editEntry] \
					\[[itcl::code $this curSelection]\]"
    bind $tree <Key-E>		[itcl::code $this changeEditor]
    bind $tree <Key-s>		"[itcl::code $this reloadHelp] \
					\[[itcl::code $this curSelection]\]"
    bind $tree <ButtonRelease-2> "
	_tixRecordFlags <ButtonRelease-2>  %% %# %b %c %d %f %h %k %m %o %p \
		%s %t %w %x %y %A %B %E %K %N %R %T %W %X %Y
	[itcl::code $this showCmd] \[menuHelp::bindButton %W %x %y\]
	_tixDeleteFlags
    "
    bind $tree <ButtonRelease-3> "
	_tixRecordFlags <ButtonRelease-3>  %% %# %b %c %d %f %h %k %m %o %p \
		%s %t %w %x %y %A %B %E %K %N %R %T %W %X %Y
	[itcl::code $this editEntry] \[menuHelp::bindButton %W %x %y\]
	_tixDeleteFlags
    "

    # Wheel mouse
    bind $tree <Button-4> "%W yview scroll -5 units ; break"
    bind $tree <Button-5> "%W yview scroll 5 units ; break"
    bind $tree <Shift-Button-4> "%W yview scroll -1 units ; break"
    bind $tree <Shift-Button-5> "%W yview scroll 1 units ; break"
    bind $tree <Control-Button-4> "%W yview scroll -1 pages ; break"
    bind $tree <Control-Button-5> "%W yview scroll 1 pages ; break"
    bind $tree <Alt-Button-4> "%W xview scroll -5 units ; break"
    bind $tree <Alt-Button-5> "%W xview scroll 5 units ; break"

    # Pack windows
    pack $b -side top -fill both -expand 0 -pady 4
    pack $b.exit -side right
    pack $b.search -side right
    pack $b.status -side left
    pack $top.search -side top -fill both -expand 0
    pack $g.search -side right
    pack $g.help -side right
    pack $g.mtime -side left
    if {[string length $glimpseDir]} {
	pack $g -side top -anchor c -fill both -expand 0
    }
    pack $top.help -side top -fill both -expand 0
    pack $top.tree -side top -fill both -expand 1

    # Show default help
    showHelp

    # Add root entry
    $tree add "." -itemtype text -text "$title"
    set _tree(.,cmd) {}
    set _tree(.,label) "$title"
    set _tree(.,helpText) "\n $title menu."

    # Convert menu into a tree
    buildTree {} $menu
    $top.tree autosetmode

    # Report time stamp of glimpse index
    glimpseMtime

    # End of constructor
    return
}

# Destructor
itcl::body menuHelp::destructor {} {
    # Unbind <Destroy>, and then destroy...
    bind $top <Destroy> {}
    destroy $top
}

# Return the version number
itcl::body menuHelp::version {} {
    return [lindex $version 1]
}

########################################################################
#                                                                      #
# User preferences                                                     #
#                                                                      #
########################################################################

# Initialize preferences.
itcl::body menuHelp::preferences {} {
    global env

    # Editor
    option add *Menuhelp.editor 	    /bks/proj/bin/nedit 20

    # A few color resources I don't like from Tix
    option add *Menuhelp*disabledForeground #a3a3a3		20
    option add *Menuhelp*activeBackground   gray95		interactive
    option add *Menuhelp*activeForeground   black		interactive
    option add *Menuhelp*troughColor	    pink		interactive

    # If the user has any preferences, they can put them here...
    catch "option readfile $env(HOME)/.menuhelp-defaults 90"
    return
}

# Allow user to change the editor.
itcl::body menuHelp::changeEditor {} {
    # Create a toplevel dialog
    destroy $top.newEdit
    toplevel $top.newEdit -class Dialog -relief ridge

    # Create a tixLabeledEntry
    tixLabelEntry $top.newEdit.entry -label "Select Editor:"
    set e [$top.newEdit.entry subwidget entry]

    #
    # Buttons
    #

    set b [frame $top.newEdit.buttons]
    # OK button
    button $b.ok -text OK -relief raised -takefocus 0 \
	-command "
	    option add *editor \"\[$e get\]\" 99
	    destroy $top.newEdit
	"
    bind $top.newEdit <Return> "
	option add *editor \"\[$e get\]\" 99
	destroy $top.newEdit
    "
    # Cancel button
    button $b.cancel -text Cancel -relief raised -takefocus 0 \
	-command "destroy $top.newEdit"

    # Use imageLib images if available
    catch "imageLib::imageConfig $b.ok OK -relief flat"
    catch "imageLib::imageConfig $b.cancel Cancel -relief flat"

    # Pack everything
    pack $top.newEdit.entry -side top -fill x -expand 1
    pack $b -side top -fill both -expand 1 -anchor center
    pack $b.ok -side right -fill both -expand 1
    pack $b.cancel -side right -fill both -expand 1

    # Default is the current editor
    $e delete 0 end
    $e insert end [option get $top editor {}]
}

# Get the current selection in the tree (if any).
itcl::body menuHelp::curSelection {} {
    return [$tree info selection]
}

########################################################################
#                                                                      #
# Static Procedures                                                    #
#                                                                      #
########################################################################

# Read the comments from the fileName.
proc menuHelp::readHelpFile {fileName} {
    # Open file and initialize input
    set file [open $fileName]
    set head {}

    # Get file extension
    set ext [translit A-Z a-z [lindex [split $fileName "."] end]]
    # Based on file type, create regular expression of what to drop
    switch -regexp -- $ext {
	^sas$ {
	    set dropRE {(/\*|\*/|\*|%\*|;)}
	}
	^(inp|pro)$ {
	    set dropRE {--}
	}
	-defaults(-color|-mono|-low|)$ {
	    set dropRE {!}
	}
	default {
	    set dropRE {#}
	}
    }

    # Ignore the first line
    gets $file

    # Read the first 20 lines, or until the first blank.
    set lastLine {}
    set j 0
    set done 0
    while {!$done} {
	# Read the next line
	gets $file nextLine
	# If the line is blank, we're done
	if {![string length $nextLine]} {
	    set done 1
	} else {
	    # If the line contains "prowish", ignore it...
	    if {[regexp {(prowish|protclsh)} $nextLine]} {
		set nextLine {}
	    }
	    # Remove comment characters
	    regsub -all -- $dropRE $nextLine {} nextLine
	    # Trim white space.
	    set nextLine [string trim $nextLine]

	    # If line begins with a dot (.), we don't show it
	    if {[regexp {^\.} $nextLine]} {
		set nextLine {.}
		incr j -1
	    }

	    if {![string length $nextLine]} {
		if {![string length $lastLine]} {
		    append head "\n"
		} else {
		    if {![string equal . $lastLine]} {
			append head "\n\n"
		    }
		}
	    } else {
		if {[string equal . $nextLine]} {
		    if {![string equal . $lastLine]} {
			append head "\n"
		    }
		    set nextLine {.}
		} else {
		    if {![string length $lastLine] || \
			    [string equal . $lastLine]} {
			append head "$nextLine"
		    } else {
			append head " $nextLine"
		    }
		}
	    }
	    # If trailing character is "$", add newline
	    if [regexp {\$$} $nextLine] {
		append head "\n"
		set nextLine {}
	    }
	}
	set lastLine $nextLine
	incr j
	if {$j >= 20} {
	    set done 1
	}
    }
    close $file

    # Remove "$" at end of lines
    regsub -all "\\$\n" $head "\n" head
    # Remove trailing newlines and return results
    regsub "\[\n\]+$" $head {} head
    # Trim whitespace
    set head [string trim $head]
    # Remove multiple new lines
    regsub -all "\n\n\n" $head "\n\n" head
    # Remove man formatting
    regsub -all {(\\fI|\\fR|\\fB)} $head {} head
    regsub -all {\*([^*]+)\*} $head {\1} head
    return $head
}

# Find which entry was clicked.
proc menuHelp::bindButton {w x y} {
    set ent [tixHList:GetNearest $w $y]

    # Try and find the entry clicked
    if {[string length $ent]} {
	set info [$w info item $x $y]
	set key [lindex $info 0]
	return $key
    } else {
	return
    }
}

########################################################################
#                                                                      #
# Methods for building the tree                                        #
#                                                                      #
########################################################################

# Loop through menu antries and add to the tree.
itcl::body menuHelp::buildTree {addPath m} {
    # Loop through menu and add entries
    loop j 1 [expr {[$m index end] +1}] {
	incr menuCnt

	# Determine entry type and add to the tree
	set t [$m type $j]
	set key $addPath.entry_$j
	switch -exact $t {
	    command {
		# Get the command and add it to the tree
		set lab [$m entrycget $j -label]
		set state [$m entrycget $j -state]
		if {[string equal $state "disabled"]} {
		    set style "$disabledStyle"
		} else {
		    set style {}
		}
		$tree add $key -itemtype text -text $lab \
			-state $state -style $style
		# Add command and, if available, help
		set cmd [$m entrycget $j -command]
		set _tree($key,label) $lab
		set _tree($key,cmd) $cmd
		assignAttributes $key
	    }
	    checkbutton {
		# Create a checkbutton entry
		regsub -all {\.} $key {_} subwindow
		set lab [$m entrycget $j -label]
		set cmd [$m entrycget $j -command]
		set v [$m entrycget $j -variable]
		set state [$m entrycget $j -state]
		checkbutton $tree.check_$subwindow -text $lab \
		    -padx 0 -pady 0 -command $cmd -variable $v \
		    -borderwidth 2 -state $state
		$tree add $key -state $state \
		    -itemtype window -window $tree.check_$subwindow
		# Store label/command and assign attributes
		set _tree($key,label) $lab
		set _tree($key,cmd) $cmd
		assignAttributes $key
	    }
	    cascade {
		# Add the submenu
		set lab [$m entrycget $j -label]
		set sub [$m entrycget $j -menu]
		$tree add $key -itemtype text -text $lab
		set _tree($key,label) $lab
		set _tree($key,cmd) {}
		# Include all ancestors in helpText...
		set indent {}
		set _tree($key,helpText) {}
		foreach k [findPath $key] {
		    append _tree($key,helpText) "\n$indent$_tree($k,label)"
		    append indent "  "
		}
		append _tree($key,helpText) " menu."
		# Build the submenu (call this method recursively)
		buildTree $key $sub
		# Hide the children from the submenu
		foreach k [$tree info children $key] {
		    $tree hide entry $k
		}
	    }
	    separator {
		# Add a separator
		$tree add $key -itemtype image -image ::imageLib::MenuSep \
		    -state disabled
		set _tree($key,label) {}
		set _tree($key,cmd) {}
		set _tree($key,helpText) {----------}
	    }
	    default {}
	}
    }
}

# Extract comments from a command.
itcl::body menuHelp::getComments {cmd} {
    set comments [lmatch -regexp [split $cmd "\n"] "^(\[ \t\]+|)#.*$"]
    if {[string length $comments]} {
	return $comments
    } else {
	return
    }
}

# Assign the help file and the edit file for key.
itcl::body menuHelp::assignAttributes {key} {
    set comments [getComments $_tree($key,cmd)]
    # If first comment is "Edit file:" then set the editFile
    # attribute to the filename and remove this comment.
    if {[string equal [lrange [lindex $comments 0] 1 2] "Edit file:"]} {
	set _tree($key,editFile) [lindex [lindex $comments 0] 3]
	set comments [lrange $comments 1 end]
	# If the next comment is "Help file:", then set the helpFile
	# attribute, otherwise helpFile is editFile...
	if {[string equal [lrange [lindex $comments 0] 1 2] "Help file:"]} {
	    set _tree($key,helpFile) \
		[lindex [lindex $comments 0] 3]
	    set comments [lrange $comments 1 end]
	} else {
	    set _tree($key,helpFile) $_tree($key,editFile)
	}
    } else {
	# No editFile or helpFile
	set _tree($key,editFile) {}
	set _tree($key,helpFile) {}
    }
    # Remove leading comment characters
    regsub -all "\n(\[ \t\]+|)#" "\n[join $comments \n]" "\n" \
	helpTxt
    set _tree($key,helpText) [string range $helpTxt 1 end]
    return
}

########################################################################
#                                                                      #
# Methods for hiding or displaying branches                            #
#                                                                      #
########################################################################

# Hide all entries.
itcl::body menuHelp::hideAll {{key {}}} {
    foreach child [$tree info children $key] {
	$tree hide entry $child
	hideAll $child
    }
    return
}

# Return all keys that lead to path.
itcl::body menuHelp::findPath {key} {
    # Initialize
    set fullPath {.}
    set path {}

    # Loop through components
    foreach k [lrange [split $key "."] 1 end] {
	append path ".$k"
	lappend fullPath $path
    }
    return $fullPath
}

# Show all keys leading to child.
itcl::body menuHelp::showKey {key} {
    foreach k [findPath $key] {
	$tree show entry $k
    }
    return
}

########################################################################
#                                                                      #
# Methods for executing or editing menu items                          #
#                                                                      #
########################################################################

# Show the command of an entry.
itcl::body menuHelp::showCmd {key} {
    if {[info exists _tree($key,cmd)]} {
	showHelp {} $_tree($key,cmd)
    }
    return
}

# Execute menu entry.
itcl::body menuHelp::runEntry {key} {
    if {[info exists _tree($key,cmd)]} {
	if {[string length $_tree($key,cmd)]} {
	    showHelp {} "\nRunning $_tree($key,label)."
	    namespace inscope :: "$_tree($key,cmd)"
	}
    }
    return
}

# Edit the editFile for an entry.
itcl::body menuHelp::editEntry {{key {}}} {
    set editFile {}

    # Set key to current selection and get the name of the editFile
    if {[string length $key]} {
	$tree selection clear
	$tree selection set $key
	$tree anchor set $key
	if {[info exists _tree($key,editFile)]} {
	    set editFile $_tree($key,editFile)
	}
    }

    # If an edit file was found, edit it
    if {[string length $editFile]} {
	set editor [option get $top editor {}]
	showHelp {} "\nEditing:\n$editFile"
	system "$editor $editFile" &
    } else {
	showHelp {} "No edit file..."
    }
    return
}

########################################################################
#                                                                      #
# Methods for the HELP/SEARCH facility                                 #
#                                                                      #
########################################################################

# Show the SEARCH help.
itcl::body menuHelp::searchHelp {} {
    showHelp {} {
Enter a regular expression to
search for and hit Return or
click "Search" to find.}
    return
}

# Get helpText for specified key.
itcl::body menuHelp::getHelp {key} {
    # Check if help already exists
    if {![info exists _tree($key,helpText)]} {
	set _tree($key,helpText) {}
    }
    # If helpText is empty, try and load it
    if {![string length $_tree($key,helpText)]} {
	# No help exists, so see if helpFile exists...
	if {[info exists _tree($key,helpFile)]} {
	    if {[string length $_tree($key,helpFile)]} {
		catch "menuHelp::readHelpFile $_tree($key,helpFile)" helpText
		set _tree($key,helpText) $helpText
		# Finally, if no help was found, set NO HELP msg...
		if {![string length $_tree($key,helpText)]} {
		    set _tree($key,helpText) "See $_tree($key,helpFile)"
		}
	    } else {
		# Help file was null...
		set _tree($key,helpText) "Sorry, no help available..."
	    }
	} else {
	    # Help file didn't exist...
	    set _tree($key,helpText) "Sorry, no help available..."
	}
    }
    return
}

# Reload the helpText.
itcl::body menuHelp::reloadHelp {key} {
    if {[info exists _tree($key,helpFile)]} {
	if {[string length $_tree($key,helpFile)]} {
	    set _tree($key,helpText) {}
	    getHelp $key
	    showHelp $key
	}
    }
}

# Show the helpText.
itcl::body menuHelp::showHelp {{key {}} {helpText {}}} {
    # If a key was specified, get help for that key
    if {[string length $key]} {
	getHelp $key
	set helpText $_tree($key,helpText)
    }
    if {![string length $helpText]} {
	set helpText {
Right-click on entry to edit.

Double-click or Return to
execute.

Click on +/- to show/hide
submenus.}}

    $help configure -state normal
    $help delete 0.0 end
    $help insert 0.0 "$helpText"
    $help configure -state disabled
    update idletasks
    return
}

# Load ALL the helpText (needed for full search).
itcl::body menuHelp::loadAllHelp {{key .}} {
    # Loop over all children of $key
    foreach child [$tree info children $key] {
	# Update progress meter
	updateMeter
	$top.meter.meter configure -text $_tree($child,label)
	update
	# Load help and call loadAllHelp recursively
	getHelp $child
	loadAllHelp $child
    }
    return
}

# Update tixMeter.
itcl::body menuHelp::updateMeter {{title Loading...}} {
    set m $top.meter
    if {![info exists meter]} {
	destroy $m
	toplevel $m -relief ridge -class Meter
	tixMeter $m.meter -fillcolor yellow -borderwidth 0
	pack $m.meter
	wm resizable $m 0 0
	wm title $m $title
	wm geometry $m +20+20
	set meter 0
    } else {
	incr meter
	$m.meter configure -value [expr "$meter.0 / $menuCnt"]
    }
    return
}

# Search help files and menu entries for regular expression
itcl::body menuHelp::apropos {} {
    # If help is not loaded, do so now...
    if {!$helpLoaded} {
	set loadNow [MessageDlg $top.loadnow \
		-separator 1 \
		-title "Load all help files now?" \
		-justify left \
		-font 10x20 \
		-icon question \
		-message {You might not find all
matches until all
help files are loaded.

This may take a while...
Load Now?} \
		-type yesno]
	grab .

	if {!$loadNow} {
	    $status configure -text "Loading help\nfiles..." -foreground red
	    configure -helpLoaded 1
	    puts stdout "Loading all help files..."
	    update idletasks
	    catch "unset meter"
	    loadAllHelp
	    catch "unset meter"
	    destroy $top.meter
	    $status configure -text "Help\nloaded..." -foreground darkgreen
	    puts stdout "done."
	    update idletasks
	}
    }

    # Flash the search button and set status
    $top.buttons.search flash
    $status configure -text "Searching...\n" -foreground darkgreen
    update idletasks

    # Get entry and begin the search
    set search [$entry get]
    set allHelp [lmatch -regexp [array names _tree] ".*,(helpText|label)"]
    set matches {}
    foreach k $allHelp {
	if {[regexp -nocase -- $search $_tree($k)]} {
	    lappend matches [lindex [split $k ","] 0]
	}
    }
    set matches [lrmdups $matches]

    # If there are matches, hide all and show matches
    if {[llength $matches] != 0} {
	$status configure -text "Found\n[llength $matches] entries." \
		-foreground darkgreen
	hideAll .
	foreach m $matches {
	    showKey $m
	}
	$top.tree autosetmode
    } else {
	$status configure -text "No matches\nfound..." -foreground red
    }
    return
}

# Search contents of help files for regular expression via egrep.
itcl::body menuHelp::egrep {} {
    # Flash the search button and set status
    $top.buttons.search flash
    $status configure -text "Grepping...\n" -foreground darkgreen
    update idletasks

    # Initialize meter
    catch "unset meter"
    # Get search pattern from entry
    set search [$entry get]
    # Locate all branches to search
    set children [lmatch -regexp [array names _tree] ".*,(label)"]
    # Begin search
    set matches {}
    foreach child $children {
	# Get just the key
	set key [lindex [split $child ","] 0]
	# Update meter
	updateMeter "Grepping..."
	$top.meter.meter configure -text $_tree($child)
	update idletasks
	# Check for help file
	if {[info exists _tree($key,helpFile)]} {
	    set fileName $_tree($key,helpFile)
	    # Check if there is a filename
	    if {[string length $fileName]} {
		# egrep file
		if {![catch "exec egrep -i \"$search\" $fileName" result]} {
		    lappend matches $key
		}
	    }
	}
    }
    # Sort matches and remove duplicates
    set matches [lrmdups $matches]
    # Clear meter
    catch "unset meter"
    destroy $top.meter

    # If there are matches, hide all and show matches
    if {[llength $matches] != 0} {
	$status configure -text "Found\n[llength $matches] entries." \
	    -foreground darkgreen
	hideAll .
	foreach m $matches {
	    showKey $m
	}
	$top.tree autosetmode
    } else {
	$status configure -text "No matches\nfound..." -foreground red
    }
    return
}

# Return a list of all helpFiles.
itcl::body menuHelp::listHelpFiles {{showKeys 0}} {
    set children [lmatch -regexp [array names _tree] ".*,(label)"]
    set matches {}
    # Begin search
    foreach child $children {
	# Get just the key
	set key [lindex [split $child ","] 0]
	# Check for help file
	if {[info exists _tree($key,helpFile)]} {
	    set fileName $_tree($key,helpFile)
	    # Check if there is a filename
	    if {[string length $fileName]} {
		if {$showKeys} {
		    lappend matches [list $fileName $key]
		} else {
		    lappend matches $fileName
		}
	    }
	}
    }
    # Sort matches and remove duplicates
    set matches [lrmdups $matches]
    # Return the matches
    return $matches
}

# Check if the glimpse index is out-dated (it exists, but not up-to-date).
itcl::body menuHelp::glimpseOutdated {} {
    if {![string length $glimpseDir]} {
	# We return 0 here because there is no glimpse directory,
	#   and therefore it is, by definition, up to date.
	return 0
    }
    # Check if the index exists and get is mtime
    if {[catch {file mtime $glimpseDir/.glimpse_index} result]} {
	# Couldn't get the mtime, so it either doesn't exist
	#   or something else is terribly wrong.  Either way,
	#   it needs to be updated.
	return 1
    } else {
	set glimpseMtime $result
    }

    # Get the helpFile names (to check mtimes)
    set helpFiles [listHelpFiles]
    # Initialize list of mtimes
    set helpMtime {}
    # Loop through files and get mtime
    foreach f $helpFiles {
	if {[catch {file mtime $f} result]} {
	    lappend helpMtime [list 0 $f]
	} else {
	    lappend helpMtime [list $result $f]
	}
    }

    # The last modified time is the last element of the sorted list
    set latestUpdated [lindex [lsort -index 0 -integer $helpMtime] end]
    set helpMtime [lindex $latestUpdated 0]
    set helpMtimeFileName [lindex $latestUpdated 1]

    # Compare glimpseMtime with helpMtime
    if {[expr {$glimpseMtime < $helpMtime}]} {
	# Glimpse appears to be out of date
	return 1
    } else {
	# Glimpse must be OK
	return 0
    }
}

# Report the modification time of the glimpse index.
itcl::body menuHelp::glimpseMtime {} {
    # Check if this has a glimpse directory...
    if {![string length $glimpseDir]} {
	return 1
    }
    # Get the mtime
    if {[catch "file mtime $glimpseDir/.glimpse_index" mtime]} {
	# Some sort of error, so must be no index
	$top.glimpse.mtime configure -text "No index" -foreground red
	return 1
    } else {
	set t [clock format $mtime -format "Indexed: %H:%M\n%m/%d/%Y"]
	set k "[expr {[file size $glimpseDir/.glimpse_index]/1000}]K"
	$top.glimpse.mtime configure -text "$t ($k)"
	# Check time stamp on glimpse index vs. help files
	if {[glimpseOutdated]} {
	    $top.glimpse.mtime configure -foreground red
	} else {
	    $top.glimpse.mtime configure -foreground darkgreen
	}
	return 0
    }
}

# Show the last modified time.
itcl::body menuHelp::helpMtime {} {
    # Check if this has a glimpse directory...
    if {![string length $glimpseDir]} {
	return 1
    }
    # Update glimpse mtime
    if {![glimpseMtime]} {
	# Update latest helpFile mtime
	set t [clock format $helpMtime -format "\n    %H:%M:%S\n    %m/%d/%Y"]
	showHelp {} "\nLast file updated:\n$t\n([lindex \
		[split $helpMtimeFileName "/"] end])"
	return 0
    } else {
	# Wasn't able to get the latest helpFile mtime
	return 1
    }
}

# Create a glimpse index of the helpFiles in user-specified directory.
itcl::body menuHelp::glimpseIndex {} {
    # Check if this has a glimpse directory...
    if {![string length $glimpseDir]} {
	return 1
    }
    # Get the helpFile names
    set helpFiles [listHelpFiles]
    # Write the file names into .glimpse_include
    #   (even though we tell it explicitly which files to index,
    #   we must add them to the include file to force it to index
    #   symbolic links)
    set includeFile [open $glimpseDir/.glimpse_include w]
    puts $includeFile [join $helpFiles "\n"]
    close $includeFile
    # Build the glimpseindex
    $status configure -text "Indexing...\n" -foreground darkgreen
    update idletasks
    system "(echo [join $helpFiles \"\n\"] |\
	/bks/proj/bin/glimpseindex -H $glimpseDir -F -n ; \
	chmod ug=rw,o=r $glimpseDir/.glimpse_*)"
    $status configure -text "Index\nBuilt" -foreground darkgreen
    glimpseMtime
    return 0
}

# Use glimpse to search.
itcl::body menuHelp::glimpseSearch {} {
    # Check if this has a glimpse directory...
    if {![string length $glimpseDir]} {
	return [apropos]
    }
    # Update glimpse time stamp and check if it is up-to-date still
    glimpseMtime

    # Flash the search button and set status
    $top.glimpse.search flash
    $status configure -text "Glimpse...\n" -foreground darkgreen
    update idletasks

    # Get the search pattern and run glimpse
    set glimpseOpts {}
    set search [$entry get]
    if {[llength $search] > 1} {
	if {[string equal [string index $search 0] "-"]} {
	    set glimpseOpts [lindex $search 0]
	    set search [lrange $search 1 end]
	}
    }
    if {[catch {exec /bin/csh -cf "
		/bks/proj/bin/glimpse -y $glimpseOpts -H $glimpseDir '$search'
	    "} result]} {
	if {[string equal {child process exited abnormally} $result]} {
	    # Linux glimpse seems to use return code 1 for no matches
	    $status configure -text "No matches\nfound..." \
		-foreground red
	} else {
	    $status configure -text "ERROR:\nsee below" -foreground red
	    showHelp {} $result
	}
    } else {
	# See if we made any matches
	if {[string length $result]} {
	    # Is there an OutWindow already?
	    set glimpseOut $top.outwindow
	    if {![winfo exists $glimpseOut]} {
		# Get resources
		toplevel $glimpseOut -class GlimpseOut
		set fn [option get $glimpseOut {} Font]
		if {![string length $fn]} {
		    set fn -*-clean-medium-r-*-*-14-*-*-*-*-*-*-*
		}
		set geom [option get $glimpseOut geometry {}]
		if {![string length $geom]} {
		    set geom "100x24+402+324"
		}
		destroy $glimpseOut
		# Build the window
		ToplevelOutput $glimpseOut \
		    -buttons "search {exit right wm withdraw $glimpseOut}" \
		    -update 0 \
		    -textfont $fn \
		    -geometry $geom \
		    -textbackground white \
		    -foreground black \
		    -title {Glimpse Results} \
		    -iconname {Glimpse}
		update idletasks
		wm withdraw $glimpseOut
		update idletasks
		wm deiconify $glimpseOut
	    } else {
		wm deiconify $glimpseOut
	    }
	    $glimpseOut clear
	    # Get the editor command
	    #   gvim supports passing a search string
	    set editor [option get $top editor {}]
	    if {[regexp "gvim|xvi" [file tail [lindex $editor 0]]]} {
		regsub -all "\"" $search {} vimsearch
		regsub -all {;} $search {\|} vimsearch
		set editCmd "-c \"set ic|/$vimsearch\""
	    } else {
		set editCmd {}
	    }
	    # Clear old tags
	    foreach delTag [$glimpseOut tag names] {
		if {[regexp {^/} $delTag]} {
		    $glimpseOut tag delete $delTag
		}
	    }
	    bind [$glimpseOut component text] <Double-1> {break}
	    # Loop through lines
	    set lastfile {}
	    foreach line [split $result "\n"] {
		set filename [lindex [split $line :] 0]
		regsub -- "^${filename}:" $line {} match
		# Pad filenames with white space
		if {![string match $lastfile $filename]} {
		    $glimpseOut puts stdout \
			"[string repeat - 20]+[string repeat - 55]"
		    set filetail [file tail $filename]
		} else {
		    set filetail {}
		}
		set strlen [string length $filetail]
		set pad [string repeat { } [expr {20-$strlen}]]
		$glimpseOut puts -nonewline stdout "$filetail$pad|"
		set lastfile $filename
		# Tag the filename
		$glimpseOut taglast $filename
		$glimpseOut tag bind $filename <Enter> "
		    $glimpseOut tag configure $filename \
			-background gray90
		    $glimpseOut tag raise $filename
		"
		$glimpseOut tag bind $filename <Leave> "
		    $glimpseOut tag configure $filename \
			-background {}
		    $glimpseOut tag lower $filename
		"
		$glimpseOut tag bind $filename <Double-1> \
		    "system {$editor $filename $editCmd} &;break"
		$glimpseOut tag raise $filename
		# Add the match
		regsub -all {;} $search {|} pattern
		regsub -all -- $pattern $match "\n\\0\n" newMatch
		foreach subLine [split $newMatch \n] {
		    $glimpseOut puts -nonewline stdout $subLine
		    $glimpseOut taglast $filename
		    if {[regexp -- $pattern $subLine]} {
			$glimpseOut colorlast darkgreen yellow
		    } else {
			$glimpseOut colorlast blue
		    }
		}
		$glimpseOut puts stdout {}
		$glimpseOut taglast $filename
	    }
	    raise $glimpseOut
	    $glimpseOut see 0.0
	    showHelp {} "\nShowing matches..."
	    update idletasks
	    # Split results into files...
	    regsub -all -- ":\[^\n]*\n" "$result\n" "\n" files
	    set files [lrmdups $files]
	    $status configure -text "Found\n[llength $files] files." \
		-foreground darkgreen
	    update idletasks
	    # Show all files in the tree
	    hideAll .
	    set helpIndex [listHelpFiles 1]
	    foreach f $files {
		set matches [lmatch -regexp $helpIndex "^$f "]
		foreach m $matches {
		    showKey [lindex $m 1]
		}
	    }
	    $top.tree autosetmode
	    showHelp
	} else {
	    $status configure -text "No matches\nfound..." \
		-foreground red
	}
    }
    return 0
}

# Give some help for Glimpse
itcl::body menuHelp::glimpseHelp {} {
    showHelp {} {Glimpse (GLobal IMPlicit SEarch) searches the\
CONTENTS (not just the header) of the Help Files using an index.

Enclose multi-word searches 'in quotes'.

Prefix search with "-i" to ignore case.
}
    return
}
